{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Image Server Node"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "from viper_toolkit import Dissect\n",
    "import inspect"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {
    "scrolled": false
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "class ImageServer(object):\n",
      "    def __init__(self):\n",
      "        self.setup_ros()\n",
      "        self.wait_for_dependencies()\n",
      "        self.loop()\n",
      "\n",
      "    def setup_ros(self):\n",
      "        # This runs all of the preceeding functions.\n",
      "        rospy.init_node('image_server', log_level=rospy.DEBUG)\n",
      "        self.setup_parameters()\n",
      "        self.setup_ticket()\n",
      "        self.setup_request_nodes()\n",
      "        self.setup_image_nodes()\n",
      "        self.setup_publishers()\n",
      "        self.setup_model_server_proxy()\n",
      "        rospy.loginfo(f\"[{self.name.abv}] ROS Setup\")\n",
      "    \n",
      "    def setup_parameters(self):\n",
      "        #Create an instance of the parameters object.\n",
      "        self.name = NameManager()\n",
      "        self.logger = Logger(name_manager = self.name) \n",
      "        self.img_s_parameters = ParameterManager(\n",
      "            logger = self.logger, \n",
      "            name_manager = self.name)\n",
      "        \n",
      "        # Add each of our parameters and indicate the rate we wish\n",
      "        # for the noze to run\n",
      "        self.img_s_parameters.add(\n",
      "            Parameter(\n",
      "                name=\"rate\", \n",
      "                target=f\"{self.name.name}/rate\", \n",
      "                default=50,\n",
      "                dynamic=True,\n",
      "                name_manager = self.name,\n",
      "                logger=self.logger))\n",
      "\n",
      "        # Indicates that the parameter should be updated dynamically \n",
      "        # if it changes after the node has started running.\n",
      "        self.img_s_parameters.add(\n",
      "            Parameter(\n",
      "                name=\"updates\", \n",
      "                target=f\"{self.name.name}/dynamic\", \n",
      "                default=True, \n",
      "                dynamic=True,\n",
      "                name_manager=self.name,\n",
      "                logger=self.logger))\n",
      "\n",
      "    def setup_ticket(self):\n",
      "        # Instantiates a ticket object which contain both\n",
      "        # the flags for the type of image requested, as well\n",
      "        # as the last image itself\n",
      "        self.ticket = Ticket()\n",
      "        rospy.loginfo(f\"[{self.name.abv}] Ticket Setup\")\n",
      "\n",
      "    def wait_for_dependencies(self):\n",
      "        # There is no functionality for the image server unless our \n",
      "        # camera module is online. \n",
      "        dependencies = [self.camera_node]\n",
      "        for package in dependencies:\n",
      "            rospy.loginfo(f\"[{self.name.abv}] Waiting for {package} to come online.\")\n",
      "            rospy.wait_for_message(package, Image)\n",
      "            \n",
      "        rospy.loginfo(f\"[{self.name.abv}] {package} Online.\")\n",
      "        return\n",
      "        \n",
      "\n",
      "        \n",
      "    def setup_request_nodes(self):\n",
      "        # Subscribe to the image request message from the Pose \n",
      "        # Estimation node\n",
      "        rospy.Subscriber('image_request/POS', \n",
      "            Bool, \n",
      "            self.image_rqst_callback,\n",
      "            queue_size = 1\n",
      "            )\n",
      "        rospy.loginfo(f\"[{self.name.abv}] Pose Detection Node Subscribed\")\n",
      "        \n",
      "        # Subscribe to the Image request message from the Scene \n",
      "        # Segmentation node\n",
      "        rospy.Subscriber('image_request/SEG', \n",
      "            Bool, \n",
      "            self.image_rqst_callback,\n",
      "            queue_size = 1\n",
      "            )\n",
      "        rospy.loginfo(f\"[{self.name.abv}] Segmentation Node Subscribed\")\n",
      "        \n",
      "        # Subscribe to the image request message from the depth \n",
      "        # perception node\n",
      "        rospy.Subscriber('image_request/DPT', \n",
      "            Bool, \n",
      "            self.image_rqst_callback,\n",
      "            queue_size = 1\n",
      "            )\n",
      "        rospy.loginfo(f\"[{self.name.abv}] Depth Perception Node Subscribed\")\n",
      "    \n",
      "    def setup_image_nodes(self):\n",
      "        # Create the holding container for an incoming image message.\n",
      "        self.image = Image()\n",
      "        \n",
      "        # Subscribe to the camera node.\n",
      "        self.camera_node = \"/inland_ir_cam/image\"\n",
      "        rospy.Subscriber(self.camera_node,\n",
      "            Image, \n",
      "            self.image_callback, \n",
      "            queue_size=1\n",
      "            )\n",
      "        rospy.loginfo(f\"[{self.name.abv}] Camera Subscribed\")\n",
      "    \n",
      "    def setup_publishers(self):\n",
      "        # This is the method by which we publish the most recent image\n",
      "        # Specifically, this is used by the final image augmentation \n",
      "        # node as the models receive their image via the service.\n",
      "        self.pub_image = rospy.Publisher(\n",
      "            'image_server/image',\n",
      "            Image,\n",
      "            queue_size=1\n",
      "            )\n",
      "        rospy.loginfo(\"[IMG] Image Republisher Online\")\n",
      "        \n",
      "        # This publishes the status of the images delivered to the \n",
      "        # Model Server. It announces to all nodes that new images are\n",
      "        # available at the model server and they may now make a new \n",
      "        # request.\n",
      "        self.pub_status = rospy.Publisher(\n",
      "            'image_server/status',\n",
      "            Bool,\n",
      "            queue_size=1\n",
      "            )\n",
      "        rospy.loginfo(\"[IMG] Image Server Status Online\")\n",
      "            \n",
      "    def setup_model_server_proxy(self):\n",
      "        # This is the method API by which the image server requests\n",
      "        # that the model server accept its image. The request message\n",
      "        # is the image, and the reply is a bool indicating success.\n",
      "        self.image_push_rqst = rospy.ServiceProxy(\n",
      "            'model_server/image_fetching_service',\n",
      "            ImageRequest\n",
      "            )\n",
      "\n",
      "    def image_callback(self, msg):\n",
      "        # This callback function saves the image message received \n",
      "        # from the camera to self.image\n",
      "        self.image = msg\n",
      "        rospy.logdebug(f\"[{self.name.abv}] Image Received.\")\n",
      "        \n",
      "    def image_rqst_callback(self, msg):\n",
      "        # If any message is received from a modeling node indicating a \n",
      "        # new image is required, this callback function sets the request\n",
      "        # ticket flag to True. This ensures that we only deliver images \n",
      "        # to the modeling server which will be used by the models.\n",
      "        node_status = msg.data\n",
      "        if node_status:\n",
      "            self.ticket.request = True\n",
      "\n",
      "    def action_loop(self):\n",
      "        self.ticket.request = True\n",
      "        if self.ticket.request:\n",
      "            self.timer.lap(\"Negotiate with Model Server\")\n",
      "            # If request is true, then nodes are requesting images. If\n",
      "            # response is false, then we only need to republish the\n",
      "            # new image.\n",
      "            try:\n",
      "                rospy.loginfo(\"[IMG] Image Fetching Service Requested.\")\n",
      "                response = self.image_push_rqst(self.image)\n",
      "            except:\n",
      "                rospy.logerr(\"[IMG] Could not send image to Model Server\")\n",
      "                # If we couldn't deliver the image to the model server\n",
      "                # then we need to keep trying\n",
      "                return\n",
      "\n",
      "            if response.status.data:\n",
      "                # If True, the model server now has the newest image.\n",
      "                # Note that currently this should always be True,\n",
      "                # however that will not be the case as we offer\n",
      "                # different sized images\n",
      "                try:\n",
      "                    # We tell all of the listening nodes they can now\n",
      "                    # use the Model Server.\n",
      "                    self.ticket.state = True\n",
      "                    #we clear our request queue\n",
      "                    self.ticket.request = False\n",
      "                    rospy.loginfo(\"[IMG] Images Delivered to Model Server\")\n",
      "                except:\n",
      "                    rospy.logerr(\"[IMG] Could not broadcast image available\")\n",
      "                    # If we could not change status, we should still\n",
      "                    # return the new image for nodes that are not\n",
      "                    # dependent on the model server\n",
      "                    return\n",
      "                self.timer.time(\n",
      "                    \"Negotiate with Model Server\", \n",
      "                    name = \"MOD\"\n",
      "                    )\n",
      "            else:\n",
      "                rospy.loginfo(\"[IMG] Model Server returned False\")\n",
      "                # If the ticket requests are set to false then the \n",
      "                # model server doesn't need a new image yet, however\n",
      "                # the other nodes which are not dependent on the model\n",
      "                # server do, so we will publish that image.\n",
      "                self.ticket.state = False\n",
      "                return\n",
      "\n",
      "    def loop(self):\n",
      "\n",
      "        while not rospy.is_shutdown():\n",
      "            print(self.img_s_parameters.rate)\n",
      "            print(self.img_s_parameters)\n",
      "            rate = rospy.Rate(self.img_s_parameters.rate)\n",
      "            # Update any of our parameters from the parameter server\n",
      "            if self.img_s_parameters.updates == True:\n",
      "                self.img_s_parameters.update()\n",
      "                \n",
      "            self.timer = ProcessTimer(abv=self.name.abv)\n",
      "            # See if we need to service the Model Server\n",
      "            self.action_loop()\n",
      "            \n",
      "            # Publish the current state of the image at the Model Server.\n",
      "            self.pub_status.publish(self.ticket.state)\n",
      "            \n",
      "            # Publish the last image we received regardless.\n",
      "            self.pub_image.publish(self.image)\n",
      "            rospy.logdebug(\"[IMG] Image Published\")\n",
      "            self.timer.time(\"total_runtime\")\n",
      "            rate.sleep()\n",
      "\n"
     ]
    }
   ],
   "source": [
    "from scripts.image_server import ImageServer\n",
    "source = inspect.getsource(ImageServer)\n",
    "print (source)"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.10"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
